@@ -1,9 +1,20 @@
 import asyncio
 import json
+import logging
 from asyncio import Queue
+from collections import Counter
 from functools import partial
 from pathlib import Path
-from typing import Any, Coroutine, Dict, Optional
+from typing import (
+    Any,
+    AsyncIterable,
+    Coroutine,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+)
+from uuid import UUID
 
 from kilroy_module_pytorch_py_sdk import (
     Codec,
@@ -36,6 +47,8 @@ from kilroy_module_huggingface.models import (
 from kilroy_module_huggingface.modules.base import HuggingfaceModule
 from kilroy_module_huggingface.tokenizer import HuggingfaceTokenizer
 
+logger = logging.getLogger(__name__)
+
 
 class ModelParams(SerializableModel):
     name: str
@@ -220,10 +233,10 @@ class RewardModelHuggingfaceModule(
     @staticmethod
     async def _build_reports() -> ReportsState:
         return ReportsState(
-            epoch_supervised_losses=[],
-            epoch_reinforced_scores=[],
-            epoch_reward_model_losses=[],
-            epoch_reward_model_scores=[],
+            step_supervised_losses=[],
+            step_reinforced_scores=[],
+            step_reward_model_losses=[],
+            step_reward_model_scores=[],
         )
 
     async def _build_default_state(self) -> State:
@@ -238,7 +251,7 @@ class RewardModelHuggingfaceModule(
             results_cache={},
             batch_size=params.batch_size,
             sample_size=params.sample_size,
-            epoch=0,
+            step=0,
             metrics=await self._build_metrics(),
             reports=await self._build_reports(),
             coroutine_queue=coroutine_queue,
@@ -304,11 +317,11 @@ class RewardModelHuggingfaceModule(
             "reward_model_schedulers_params": state.reward_model.schedulers_params,
             "batch_size": state.batch_size,
             "sample_size": state.sample_size,
-            "epoch": state.epoch,
-            "epoch_supervised_losses": state.reports.epoch_supervised_losses,
-            "epoch_reinforced_scores": state.reports.epoch_reinforced_scores,
-            "epoch_reward_model_losses": state.reports.epoch_reward_model_losses,
-            "epoch_reward_model_scores": state.reports.epoch_reward_model_scores,
+            "step": state.step,
+            "step_supervised_losses": state.reports.step_supervised_losses,
+            "step_reinforced_scores": state.reports.step_reinforced_scores,
+            "step_reward_model_losses": state.reports.step_reward_model_losses,
+            "step_reward_model_scores": state.reports.step_reward_model_scores,
         }
 
     @staticmethod
@@ -362,6 +375,9 @@ class RewardModelHuggingfaceModule(
                 model,
             ),
             parameters=model.parameters(),
+            **params.language_model_params.optimizers_params.get(
+                params.language_model_params.optimizer_type, {}
+            ),
         )
 
     @classmethod
@@ -382,7 +398,7 @@ class RewardModelHuggingfaceModule(
         params: Params,
         optimizer: Optimizer,
     ) -> Optional[Scheduler]:
-        if state_dict["language_model_scheduler_type"] is None:
+        if state_dict.get("language_model_scheduler_type") is None:
             return None
         return await cls._load_generic(
             directory / "language_model" / "scheduler",
@@ -394,6 +410,9 @@ class RewardModelHuggingfaceModule(
                 optimizer,
             ),
             optimizer=await optimizer.get(),
+            **params.language_model_params.schedulers_params.get(
+                params.language_model_params.scheduler_type, {}
+            ),
         )
 
     @classmethod
@@ -445,6 +464,9 @@ class RewardModelHuggingfaceModule(
                 model,
             ),
             parameters=model.parameters(),
+            **params.reward_model_params.optimizers_params.get(
+                params.reward_model_params.optimizer_type, {}
+            ),
         )
 
     @classmethod
@@ -465,7 +487,7 @@ class RewardModelHuggingfaceModule(
         params: Params,
         optimizer: Optimizer,
     ) -> Optional[Scheduler]:
-        if state_dict["reward_model_scheduler_type"] is None:
+        if state_dict.get("reward_model_scheduler_type") is None:
             return None
         return await cls._load_generic(
             directory / "reward_model" / "scheduler",
@@ -477,6 +499,9 @@ class RewardModelHuggingfaceModule(
                 optimizer,
             ),
             optimizer=await optimizer.get(),
+            **params.reward_model_params.schedulers_params.get(
+                params.reward_model_params.scheduler_type, {}
+            ),
         )
 
     @classmethod
@@ -508,6 +533,7 @@ class RewardModelHuggingfaceModule(
             directory / "frontend_generator",
             Generator,
             default=partial(cls._build_frontend_generator, params),
+            **params.frontend_generator_params,
         )
 
     @classmethod
@@ -518,6 +544,7 @@ class RewardModelHuggingfaceModule(
             directory / "backend_generator",
             Generator,
             default=partial(cls._build_backend_generator, params),
+            **params.backend_generator_params,
         )
 
     @classmethod
@@ -526,6 +553,7 @@ class RewardModelHuggingfaceModule(
             directory / "codec",
             Codec,
             default=partial(cls._build_codec, params),
+            **params.codec_params,
         )
 
     async def _load_saved_state(self, directory: Path) -> State:
@@ -549,16 +577,16 @@ class RewardModelHuggingfaceModule(
             results_cache={},
             batch_size=state_dict["batch_size"],
             sample_size=state_dict["sample_size"],
-            epoch=state_dict["epoch"],
+            step=state_dict["step"],
             metrics=await self._build_metrics(),
             reports=ReportsState(
-                epoch_supervised_losses=state_dict["epoch_supervised_losses"],
-                epoch_reinforced_scores=state_dict["epoch_reinforced_scores"],
-                epoch_reward_model_losses=state_dict[
-                    "epoch_reward_model_losses"
+                step_supervised_losses=state_dict["step_supervised_losses"],
+                step_reinforced_scores=state_dict["step_reinforced_scores"],
+                step_reward_model_losses=state_dict[
+                    "step_reward_model_losses"
                 ],
-                epoch_reward_model_scores=state_dict[
-                    "epoch_reward_model_scores"
+                step_reward_model_scores=state_dict[
+                    "step_reward_model_scores"
                 ],
             ),
             coroutine_queue=coroutine_queue,
@@ -570,3 +598,47 @@ class RewardModelHuggingfaceModule(
         while True:
             coroutine = await queue.get()
             await coroutine
+
+    async def generate(
+        self, n: int, dry: bool
+    ) -> AsyncIterable[Tuple[UUID, Dict[str, Any]]]:
+        logger.info(f"Generating {n} posts...")
+
+        async for post_id, post in super().generate(n, dry):
+            logger.info(f"Generated post {str(post_id)}.")
+            yield post_id, post
+
+        logger.info("Finished generating posts.")
+
+    async def fit_posts(
+        self, posts: AsyncIterable[Tuple[Dict[str, Any], float]]
+    ) -> None:
+        logger.info("Fitting posts...")
+
+        posts_counter = Counter(posts=0)
+
+        async def count_posts(
+            _posts: AsyncIterable[Tuple[Dict[str, Any], float]],
+            _posts_counter: Counter,
+        ) -> AsyncIterable[Tuple[Dict[str, Any], float]]:
+            async for post, score in _posts:
+                _posts_counter["posts"] += 1
+                yield post, score
+
+        await super().fit_posts(count_posts(posts, posts_counter))
+        logger.info(f"Finished fitting {posts_counter['posts']} posts.")
+
+    async def _fit_with_reward_model(self) -> None:
+        logger.info("Fitting with reward model...")
+        await super()._fit_with_reward_model()
+        logger.info("Finished fitting with reward model.")
+
+    async def fit_scores(self, scores: List[Tuple[UUID, float]]) -> None:
+        logger.info(f"Fitting {len(scores)} scores...")
+        await super().fit_scores(scores)
+        logger.info("Finished fitting scores.")
+
+    async def step(self) -> None:
+        logger.info("Performing step...")
+        await super().step()
+        logger.info("Finished performing step.")
